/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0; you may not use this file except in compliance with the Elastic License
 * 2.0.
 */

/**
 * Changelog:
 *  * v1.3 - Fixed bug that v1.2 introduced for objects with references
 *  * v1.2 - Now supports objects of any namespaceType
 *  * v1.1 - Now supports objects with references
 *  * v1.0 - Initial release
 */

const console = require('console');
const fs = require('fs');
const readline = require('readline');
const rl = readline.createInterface({
  input: process.stdin,
  output: process.stdout
});

const filePath = process.argv[2];
if (!filePath) {
  console.error('Error: Specify a file path to a valid saved object export, for example:\n  node generate_resolve_staging_data.js "/Users/joe/Downloads/export.ndjson"');
  process.exit(1);
}

const STAGING_DATA_FILE = 'resolve-staging-data.txt';
const CURL_PREFIX = 'curl -sSu elastic:changeme http://localhost:5601';
const SPACE_OBJECTS = {
  exactMatch: {
    type: 'space',
    id: 'test1',
    attributes: {
      name: 'Test 1: Resolve Exact Match',
      initials: 'T1',
      description: 'In this space, test object(s) will resolve with an "exactMatch" outcome.',
      disabledFeatures: [],
    },
  },
  aliasMatch: {
    type: 'space',
    id: 'test2',
    attributes: {
      name: 'Test 2: Resolve Alias Match',
      initials: 'T2',
      description: 'In this space, test object(s) will resolve with an "aliasMatch" outcome.',
      disabledFeatures: [],
    },
  },
  conflict: {
    type: 'space',
    id: 'test3',
    attributes: {
      name: 'Test 3: Resolve Conflict',
      initials: 'T3',
      description: 'In this space, test object(s) will resolve with a "conflict" outcome.',
      disabledFeatures: [],
    },
  },
};

async function processLineByLine(input) {
  const rl = readline.createInterface({ input });

  const objects = [];
  for await (const line of rl) {
    try {
      const obj = JSON.parse(line);
      if (obj.id) {
        objects.push(obj);
      }
    } catch (err) {
      // do nothing
    }
  }
  return objects;
}

function createAliasFor(obj) {
  const { namespace, type, id } = obj; // 'id' is the old ID
  return {
    type: 'legacy-url-alias',
    id: `${namespace}:${type}:${id}`, // the ID for the alias itself is well-formed and based upon the raw ID of the target object
    attributes: {
      targetNamespace: namespace,
      targetType: type,
      targetId: `${id}-newId`,
      sourceId: id,
    },
    references: [],
  };
}

function getBulkIndexStrings(obj) {
  const { namespace, type } = obj;
  const { attributes, id, version, ...otherObjFields } = obj;
  const action = { index: { _id: namespace ? `${namespace}:${type}:${id}` : `${type}:${id}` } };
  const doc = { [type]: attributes, ...otherObjFields };
  return [JSON.stringify(action), JSON.stringify(doc)];
}

function getKey(obj) {
  const { type, id } = obj;
  return `${type}:${id}`;
}

async function promptForNamespaceType(objType) {
  return new Promise((resolve) => {
    const prompt = `How is the "${objType}" object type CURRENTLY registered?
  [1] namespaceType: 'single' (default)
  [2] namespaceType: 'multiple-isolated'
  [3] namespaceType: 'multiple'
  [4] namespaceType: 'agnostic'
`;
    rl.question(prompt, (option) => {
      switch (option) {
        case '':
        case '1': resolve('single'); return;
        case '2': resolve('multiple-isolated'); return;
        case '3': resolve('multiple'); return;
        case '4': resolve('agnostic'); return;
      }
      console.log(`"${option}" is not a valid option, try again.`);
      resolve(promptForNamespaceType(objType));
    })
  })
}

(async function main() {
  let objects;
  try {
    const fileStream = fs.createReadStream(filePath);
    fileStream.on('error', (err) => {
      console.error('Error: Could not read file stream', err);
      process.exit(2);
    });
    objects = await processLineByLine(fileStream);
  } catch (err) {
    console.error('Error: Could not process file', err);
    process.exit(3);
  }

  if (!objects.length) {
    console.error('Error: Could not read any JSON objects! (Are you using a valid export.ndjson file from Kibana?)');
    process.exit(4);
  }

  console.log();
  console.log(`Note, this script will generate staging data to allow you to exercise the three different SavedObjectsClient.resolve() outcomes. HOWEVER, this is NOT a 100% realistic representation of objects that have been converted to become share-capable. When you implement the actual conversion, you should write integration tests to ensure that objects are migrated as expected.`);
  console.log();

  const stagingData = [];
  const objNamespaceTypes = await objects.reduce(async (acc, { type: objType }) => {
    const map = await acc;
    if (!map.has(objType)) {
      const namespaceType = await promptForNamespaceType(objType);
      map.set(objType, namespaceType);
    }
    return map;
  }, new Map());
  const objKeys = objects.reduce((acc, cur) => acc.add(getKey(cur)), new Set()); // find all objects that are included
  for (const obj of objects) {
    const { id, type, references = [] } = obj;
    const namespaceType = objNamespaceTypes.get(type);
    if (namespaceType === 'agnostic') {
      stagingData.push(obj);
      continue;
    }

    function getNsAttr(namespace) {
      if (namespaceType === 'single') {
        return { namespace };
      }
      return { namespaces: [namespace] }; // namespaceType: 'multiple' | 'multiple-isolated'
    }

    const exactMatch = {
      ...obj, ...getNsAttr(SPACE_OBJECTS.exactMatch.id), id: `${id}-test1`,
      references: references.map((ref) => {
        return (objKeys.has(getKey(ref))) ? { ...ref, id: `${ref.id}-test1` } : ref;
      })
    };
    const aliasMatch = {
      ...obj, ...getNsAttr(SPACE_OBJECTS.aliasMatch.id), id: `${id}-test2-newId`,
      references: references.map((ref) => {
        return (objKeys.has(getKey(ref))) ? { ...ref, id: `${ref.id}-test2-newId` } : ref;
      })
    };
    const conflict1 = {
      ...obj, ...getNsAttr(SPACE_OBJECTS.conflict.id), id: `${id}-test3`,
      references: references.map((ref) => {
        return (objKeys.has(getKey(ref))) ? { ...ref, id: `${ref.id}-test3` } : ref;
      })
    };
    const conflict2 = {
      ...obj, ...getNsAttr(SPACE_OBJECTS.conflict.id), id: `${id}-test3-newId`,
      references: references.map((ref) => {
        return (objKeys.has(getKey(ref))) ? { ...ref, id: `${ref.id}-test3-newId` } : ref;
      })
    };
    const aliasMatchAlias = createAliasFor({ namespace: SPACE_OBJECTS.aliasMatch.id, type: obj.type, id: `${id}-test2` });
    const conflictAlias = createAliasFor({ namespace: SPACE_OBJECTS.conflict.id, type: obj.type, id: `${id}-test3` });

    const newObjects = [exactMatch, aliasMatch, conflict1, conflict2];
    const aliases = [aliasMatchAlias, conflictAlias];

    stagingData.push(...newObjects, ...aliases);
  }
  stagingData.push(SPACE_OBJECTS.exactMatch, SPACE_OBJECTS.aliasMatch, SPACE_OBJECTS.conflict);
  const stagingDataStrings = ['POST .kibana/_bulk', ...stagingData.map((obj) => getBulkIndexStrings(obj)).flat()];

  fs.writeFileSync(STAGING_DATA_FILE, stagingDataStrings.join('\n'));

  console.log('-'.repeat(20));
  console.log(`Wrote staging data to "${process.cwd()}/${STAGING_DATA_FILE}", use Kibana Dev Tools to add it to your instance! Example:`);
  console.log(`  ${stagingDataStrings[0]}`);
  console.log(`  ${stagingDataStrings[1]}`);
  console.log(`  ...`);
  console.log();

  console.log('After loading staging data, test the different outcomes using the following CURL commands:');
  const lastObj = objects[objects.length - 1];
  const api = `/api/saved_objects/resolve/${lastObj.type}/${lastObj.id}`;
  console.log(`  ${CURL_PREFIX}/s/${SPACE_OBJECTS.exactMatch.id}${api}-test1 | npx -q json`);
  console.log(`  ${CURL_PREFIX}/s/${SPACE_OBJECTS.aliasMatch.id}${api}-test2 | npx -q json`);
  console.log(`  ${CURL_PREFIX}/s/${SPACE_OBJECTS.conflict.id}${api}-test3 | npx -q json`);
  console.log('(note, URLs need to be modified if you are running Kibana with a basePath)');
  console.log('-'.repeat(20));
})().catch((err) => {
  console.error('Error:', err);
  process.exit(5);
}).finally(() => {
  rl.close();
});
